#    Prana RC
#    Copyright (C) 2020 Dmitry Berezovsky
#    
#    prana is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#    
#    prana is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#    
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.

import argparse
import sys
import traceback
from asyncio import AbstractEventLoop

from prana_rc.service import PranaDeviceManager


class CLI:
    verbose_mode = False

    @classmethod
    def print_data(cls, msg: str):
        print(msg)

    @classmethod
    def print_info(cls, msg: str):
        print(msg, file=sys.stderr)

    @classmethod
    def print_error(cls, exception):
        print(" -> ERROR: " + str(exception), file=sys.stderr)
        if cls.verbose_mode:
            print('--------------')
            traceback.print_exc(file=sys.stderr)

    @classmethod
    def print_debug(cls, string_to_print):
        if cls.verbose_mode:
            print('[DEBUG] ' + string_to_print, file=sys.stderr)


class CliExtension(object):
    """
    Allows to extend CLI interface
    """
    COMMAND_NAME = None
    COMMAND_DESCRIPTION = None

    def __init__(self, parser: argparse.ArgumentParser, device_manager: PranaDeviceManager, loop: AbstractEventLoop):
        super().__init__()
        self.parser = parser
        self.__devce_manager = device_manager
        self.__loop = loop

    @property
    def device_manager(self) -> PranaDeviceManager:
        return self.__devce_manager

    @property
    def loop(self):
        return self.__loop

    @classmethod
    def setup_parser(cls, parser: argparse.ArgumentParser):
        pass

    def handle(self, args):
        raise NotImplementedError()


def register_global_arguments(parser: argparse.ArgumentParser):
    parser.add_argument("-v", "--verbose", dest="verbose", action='store_true', required=False,
                        help="If set more verbose output will used",
                        default=False)
    parser.add_argument("-i", "--iface", dest="iface", action='store_true', required=False,
                        help="Bluetooth interface to be used",
                        default='hci0')



#    Prana RC
#    Copyright (C) 2020 Dmitry Berezovsky
#    
#    prana is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#    
#    prana is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#    
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.

import asyncio
import bleak

from prana_rc.entity import PranaState, PranaDeviceInfo
from asyncio import AbstractEventLoop
from typing import Dict, List, Union, Optional


class PranaDeviceManager(object):
    PRANA_DEVICE_NAME_PREFIX = 'PRNAQaq'

    def __init__(self, iface: str = 'hci0', loop: Optional[AbstractEventLoop] = None) -> None:
        self.__ble_interface = iface
        self.__loop = loop
        self.__managed_devices = {}  # type: Dict['str', PranaDevice]

    @classmethod
    def __is_prana_device(cls, dev: "bleak.backends.device.BLEDevice"):
        return dev.name and dev.name.startswith(cls.PRANA_DEVICE_NAME_PREFIX)

    @classmethod
    def __prana_dev_name_2_name(cls, dev_name: str):
        return dev_name.replace(cls.PRANA_DEVICE_NAME_PREFIX, '').strip() if dev_name else dev_name

    @classmethod
    def __addr_for_target(cls, target: Union[str, PranaDeviceInfo]) -> str:
        if isinstance(target, PranaDeviceInfo):
            address = target.address
        elif isinstance(target, str):
            address = target
        else:
            raise ValueError(
                'Device must be specified either by mac address or by PranaDeviceInfo instance')
        return address

    async def discover(self, timeout: int = 5) -> List[PranaDeviceInfo]:
        """
        Listens to devices advertisement for TIMEOUT seconds and returns the list of discovered devices
        :param timeout: time to wait for devices in seconds
        :return: list of discovered devices
        """
        discovered_devs = await bleak.discover(timeout, self.__loop)
        return list(map(lambda dev: PranaDeviceInfo(address=dev.address, bt_device_name=(dev.name or '').strip(),
                                                    name=self.__prana_dev_name_2_name(dev.name), rssi=dev.rssi),
                        filter(PranaDeviceManager.__is_prana_device, discovered_devs)
                        ))

    async def connect(self, target: Union[str, PranaDeviceInfo], timeout: float = 5) -> 'PranaDevice':
        address = self.__addr_for_target(target)
        device = self.__managed_devices.get(address, None)
        if device is None:  # If not found in managed devices list
            device = PranaDevice(address, self.__loop, self.__ble_interface)
            self.__managed_devices[address] = device
        # if not await device.is_connected():
        await device.connect(timeout)
        return device

    async def disconnect_all(self):
        for dev in self.__managed_devices.values():
            await dev.disconnect()


class PranaDevice(object):
    CONTROL_SERVICE_UUID = '0000baba-0000-1000-8000-00805f9b34fb'
    CONTROL_RW_CHARACTERISTIC_UUID = '0000cccc-0000-1000-8000-00805f9b34fb'
    STATE_MSG_PREFIX = b'\xbe\xef'

    class Cmd:
        ENABLE_HIGH_SPEED = bytearray([0xBE, 0xEF, 0x04, 0x07])
        ENABLE_NIGHT_MODE = bytearray([0xBE, 0xEF, 0x04, 0x06])
        TOGGLE_FLOW_LOCK = bytearray([0xBE, 0xEF, 0x04, 0x09])
        TOGGLE_HEATING = bytearray([0xBE, 0xEF, 0x04, 0x05])
        TOGGLE_WINTER_MODE = bytearray([0xBE, 0xEF, 0x04, 0x16])

        SPEED_UP = bytearray([0xBE, 0xEF, 0x04, 0x0C])
        SPEED_DOWN = bytearray([0xBE, 0xEF, 0x04, 0x0B])
        SPEED_IN_UP = bytearray([0xBE, 0xEF, 0x04, 0x0E])
        SPEED_IN_DOWN = bytearray([0xBE, 0xEF, 0x04, 0x0F])
        SPEED_OUT_UP = bytearray([0xBE, 0xEF, 0x04, 0x11])
        SPEED_OUT_DOWN = bytearray([0xBE, 0xEF, 0x04, 0x12])

        FLOW_IN_OFF = bytearray([0xBE, 0xEF, 0x04, 0x0D])
        FLOW_OUT_OFF = bytearray([0xBE, 0xEF, 0x04, 0x10])

        STOP = bytearray([0xBE, 0xEF, 0x04, 0x01])
        READ_STATE = bytearray([0xBE, 0xEF, 0x05, 0x01, 0x00, 0x00, 0x00, 0x00, 0x5A])
        READ_DEVICE_DETAILS = bytearray([0xBE, 0xEF, 0x05, 0x02, 0x00, 0x00, 0x00, 0x00, 0x5A])

    def __init__(self, target: Union[str, PranaDeviceInfo], loop: Optional[AbstractEventLoop] = None,
                 iface: str = 'hci0') -> None:
        self.__address = None
        if isinstance(target, PranaDeviceInfo):
            self.__address = target.address
        elif isinstance(target, str):
            self.__address = target
        else:
            raise ValueError(
                'PranaDevice constructor error: Target must be eithermac address or PranaDeviceInfo instance')
        self.__client = bleak.BleakClient(self.__address, loop, device=iface)
        self.__has_connect_attempts = False

    async def __verify_connected(self):
        if not await self.is_connected():
            raise RuntimeError('Illegal state: device must be connected before running any commands')

    def notification_handler(self, sender, data):
        """Simple notification handler which prints the data received."""
        # print("{0}: {1}".format(sender, data))
        print(self.__parse_state(data))

    async def connect(self, timeout: float = 2):
        # if not await self.is_connected():
        await self.__client.connect(timeout=timeout)
        self.__has_connect_attempts = True
        await self.__client.start_notify(self.CONTROL_RW_CHARACTERISTIC_UUID, self.notification_handler)
        # TODO: Verify prana service exists to ensure it is prana device

    async def disconnect(self):
        await self.__client.disconnect()

    async def is_connected(self):
        if not self.__has_connect_attempts:
            return False
        return await self.__client.is_connected()

    async def _send_command(self, command: bytearray, expect_reply=False) -> bytearray:
        await self.__client.write_gatt_char(self.CONTROL_RW_CHARACTERISTIC_UUID, command, response=True)
        await asyncio.sleep(0.6)
        result = await self.__client.read_gatt_char(self.CONTROL_RW_CHARACTERISTIC_UUID, use_cached=False)
        if expect_reply:
            return result

    async def set_high_speed(self):
        await self.__verify_connected()
        await self._send_command(self.Cmd.ENABLE_HIGH_SPEED)

    async def set_night_mode(self):
        await self.__verify_connected()
        await self._send_command(self.Cmd.ENABLE_NIGHT_MODE)

    async def turn_off(self):
        await self.__verify_connected()
        await self._send_command(self.Cmd.STOP)

    def __parse_state(self, data: bytearray):
        if not data[:2] == self.STATE_MSG_PREFIX:
            return None
        s = PranaState()
        print(data)
        s.speed_locked = int(data[26] / 10)
        s.speed_in = int(data[30] / 10)
        s.speed_out = int(data[34] / 10)
        s.auto_mode = bool(data[20])
        s.night_mode = bool(data[16])
        s.flows_locked = bool(data[22])
        s.is_on = bool(data[10])
        s.mini_heating_enabled = bool(data[14])
        s.winter_mode_enabled = bool(data[42])
        s.is_input_fan_on = bool(data[28])
        s.is_output_fan_on = bool(data[32])
        return s

    async def read_state(self) -> PranaState:
        await self.__verify_connected()
        state_bin = await self._send_command(self.Cmd.READ_STATE, expect_reply=True)
        return self.__parse_state(state_bin)

    async def test_retrieve_state(self):
        async with bleak.BleakClient(self.__address) as client:
            self.__client = client
            return await self.read_state()



# -*- coding: utf-8 -*-

#    Prana RC
#    Copyright (C) 2020 Dmitry Berezovsky
#    
#    prana is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#    
#    prana is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#    
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Top level package for Prana RC."""

__author__ = "Dmitry Berezovsky"
__email__ = ""

import logging
import os
import sys
from prana_rc import cli

_logger = logging.getLogger(__name__)
_logger.addHandler(logging.NullHandler())
_logger.setLevel(logging.DEBUG)
if bool(os.environ.get("PRANA_LOGGING", False)):
    FORMAT = "%(asctime)-15s %(name)-8s %(levelname)s: %(message)s"
    handler = logging.StreamHandler(sys.stdout)
    handler.setLevel(logging.DEBUG)
    handler.setFormatter(logging.Formatter(fmt=FORMAT))
    _logger.addHandler(handler)

if __name__ == "__main__":
    cli.run_cli()



#    Prana RC
#    Copyright (C) 2020 Dmitry Berezovsky
#    
#    prana is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#    
#    prana is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#    
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.

from typing import NamedTuple


class PranaDeviceInfo(NamedTuple):
    address: str
    bt_device_name: str
    name: str
    rssi: int


class PranaState(object):

    def __init__(self) -> None:
        self.speed_locked: int = None
        self.speed_in: int = None
        self.speed_out: int = None
        self.night_mode: bool = None
        self.auto_mode: bool = None
        self.flows_locked: bool = None
        self.is_on: bool = None
        self.mini_heating_enabled: bool = None
        self.winter_mode_enabled: bool = None
        self.is_input_fan_on: bool = None
        self.is_output_fan_on: bool = None

    @property
    def speed(self):
        if not self.is_on:
            return 0
        return self.speed_locked if self.flows_locked else int((self.speed_in + self.speed_out) / 2)

    def __repr__(self):
        return "Prana state: {}, Speed: {}, Winter Mode: {}, Heating: {}, Flows locked: {}".format(
            'RUNNING' if self.is_on else 'IDLE',
            self.speed, self.winter_mode_enabled, self.mini_heating_enabled, self.flows_locked
        )



#    Prana RC
#    Copyright (C) 2020 Dmitry Berezovsky
#    
#    prana is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#    
#    prana is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#    
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.

import argparse
import asyncio
import logging
from typing import Type

from prana_rc.cli_utils import CliExtension, register_global_arguments, CLI
from prana_rc.service import PranaDeviceManager

supplementary_parser = argparse.ArgumentParser(add_help=False)
register_global_arguments(supplementary_parser)

parser = argparse.ArgumentParser(prog='prana', add_help=True,
                                 description='CLI interface to manage Prana recuperators via BLE',
                                 formatter_class=argparse.RawDescriptionHelpFormatter, epilog="""
--------------------------------------------------------------
Prana RC CLI Copyright (C) 2020 Dmitry Berezovsky
This program comes with ABSOLUTELY NO WARRANTY;
This is free software, and you are welcome to redistribute it
under certain conditions;
--------------------------------------------------------------
"""
                                 )
register_global_arguments(parser)
command_parser = parser.add_subparsers(title="command",
                                       dest="cmd", required=True,
                                       description='Use \"<command> -h\" to get information '
                                                   'about particular command')


# =========== AVAILABLE COMMANDS =====================

class DiscoveryCLIExtension(CliExtension):
    COMMAND_NAME = 'discover'
    COMMAND_DESCRIPTION = 'Run discovery process to find available Prana devices nearby'

    @classmethod
    def setup_parser(cls, parser: argparse.ArgumentParser):
        parser.add_argument("-t", "--timeout", dest="timeout", action='store_true', required=False,
                            help="Time in seconds to wait for devices",
                            default=5)

    async def handle(self, args):
        CLI.print_info("Starting discovery (timeout {}s)...".format(args.timeout))
        devs = await self.device_manager.discover()
        if len(devs) == 0:
            CLI.print_info("No devices found")
        else:
            for dev in devs:
                CLI.print_info(
                    "{} [{}] (identity: {}, rssi: {})".format(dev.name, dev.address, dev.bt_device_name, dev.rssi))


# =========== END OF COMMAND DEFINITIONS ==============

def read_global_args():
    known, unknown = supplementary_parser.parse_known_args()
    return known


def configure_subparser_for_cli_extension(ext: Type[CliExtension], parser: argparse.ArgumentParser,
                                          device_manager: PranaDeviceManager, loop: asyncio.AbstractEventLoop):
    ext.setup_parser(parser)
    ext_instance = ext(parser, device_manager, loop)
    parser.set_defaults(handler=ext_instance)


def run_cli():
    # logging.basicConfig(level=logging.ERROR)
    root_commands = (DiscoveryCLIExtension,)
    global_args = read_global_args()
    CLI.verbose_mode = global_args.verbose

    loop = asyncio.get_event_loop()
    device_manager = PranaDeviceManager(iface=global_args.iface)

    # Process root commands
    for cmd in root_commands:
        cmd_subparser = command_parser.add_parser(cmd.COMMAND_NAME, help=cmd.COMMAND_DESCRIPTION)
        configure_subparser_for_cli_extension(cmd, cmd_subparser, device_manager, loop)

    args = parser.parse_args()
    if hasattr(args, "handler"):
        loop.run_until_complete(asyncio.ensure_future(args.handler.handle(args)))


if __name__ == "__main__":
    run_cli()
